/*
 * bsc1173663_gem_gtt.c
 *
 * Fix for CVE-2019-0155, bsc#1173663 (i915_gem_gtt.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)
#include "bsc1173663_common.h"
#include "../kallsyms_relocs.h"

#include <linux/slab.h> /* fault-inject.h is not standalone! */
#include <linux/fault-inject.h>
#include <linux/log2.h>
#include <linux/random.h>
#include <linux/seq_file.h>
#include <linux/stop_machine.h>

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.c */
static void (*klpe_orig_gen8_ppgtt_insert_3lvl)(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   u32 unused);

static void klpp_gen8_ppgtt_insert_3lvl(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   u32 flags);

static void (*klpe_orig_gen8_ppgtt_insert_4lvl)(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   u32 unused);

static void klpp_gen8_ppgtt_insert_4lvl(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   u32 flags);

int klpp_ppgtt_bind_vma(struct i915_vma *vma,
			  enum i915_cache_level cache_level,
			  u32 unused)
{
	u32 pte_flags;
	int ret;

	if (!(vma->flags & I915_VMA_LOCAL_BIND)) {
		ret = vma->vm->allocate_va_range(vma->vm, vma->node.start,
						 vma->size);
		if (ret)
			return ret;
	}

	vma->pages = vma->obj->mm.pages;

	/* Currently applicable only to VLV */
	pte_flags = 0;
	if (vma->obj->gt_ro)
		pte_flags |= PTE_READ_ONLY;

	/*
	 * Fix CVE-2019-0155
	 *  +5 lines
	 * Double check that the callback is what it's expected to be.
	 * Note that the patched callchain should only get
	 * invoked when ->insert_entries == gen8_ppgtt_insert_[34]lvl.
	 */
	if (vma->vm->insert_entries == klpe_orig_gen8_ppgtt_insert_3lvl)
		klpp_gen8_ppgtt_insert_3lvl(vma->vm, vma, cache_level, pte_flags);
	else if (vma->vm->insert_entries == klpe_orig_gen8_ppgtt_insert_4lvl)
		klpp_gen8_ppgtt_insert_4lvl(vma->vm, vma, cache_level, pte_flags);
	else
		vma->vm->insert_entries(vma->vm, vma, cache_level, pte_flags);

	return 0;
}

static gen8_pte_t klpp_gen8_pte_encode(dma_addr_t addr,
				  /*
				   * Fix CVE-2019-0155
				   *  -1 line, +2 lines
				   */
				  enum i915_cache_level level,
				  u32 flags)
{
	/*
	 * Fix CVE-2019-0155
	 *  -2 lines, +4 lines
	 */
	gen8_pte_t pte = addr | _PAGE_PRESENT | _PAGE_RW;

	if (unlikely(flags & PTE_READ_ONLY))
		pte &= ~_PAGE_RW;

	switch (level) {
	case I915_CACHE_NONE:
		pte |= PPAT_UNCACHED;
		break;
	case I915_CACHE_WT:
		pte |= PPAT_DISPLAY_ELLC;
		break;
	default:
		pte |= PPAT_CACHED;
		break;
	}

	return pte;
}

#define kmap_atomic_px(px) kmap_atomic(px_base(px)->page)

static inline struct sgt_dma {
	struct scatterlist *sg;
	dma_addr_t dma, max;
} sgt_dma(struct i915_vma *vma) {
	struct scatterlist *sg = vma->pages->sgl;
	dma_addr_t addr = sg_dma_address(sg);
	return (struct sgt_dma) { sg, addr, addr + sg->length };
}

struct gen8_insert_pte {
	u16 pml4e;
	u16 pdpe;
	u16 pde;
	u16 pte;
};

static __always_inline struct gen8_insert_pte gen8_insert_pte(u64 start)
{
	return (struct gen8_insert_pte) {
		 gen8_pml4e_index(start),
		 gen8_pdpe_index(start),
		 gen8_pde_index(start),
		 gen8_pte_index(start),
	};
}

static __always_inline bool
klpp_gen8_ppgtt_insert_pte_entries(struct i915_hw_ppgtt *ppgtt,
			      struct i915_page_directory_pointer *pdp,
			      struct sgt_dma *iter,
			      struct gen8_insert_pte *idx,
			      /*
			       * Fix CVE-2019-0155
			       *  -1 line, +2 lines
			       */
			      enum i915_cache_level cache_level,
			      u32 flags)
{
	struct i915_page_directory *pd;
	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +1 line
	 */
	const gen8_pte_t pte_encode = klpp_gen8_pte_encode(0, cache_level, flags);
	gen8_pte_t *vaddr;
	bool ret;

	GEM_BUG_ON(idx->pdpe >= i915_pdpes_per_pdp(&ppgtt->base));
	pd = pdp->page_directory[idx->pdpe];
	vaddr = kmap_atomic_px(pd->page_table[idx->pde]);
	do {
		vaddr[idx->pte] = pte_encode | iter->dma;

		iter->dma += PAGE_SIZE;
		if (iter->dma >= iter->max) {
			iter->sg = __sg_next(iter->sg);
			if (!iter->sg) {
				ret = false;
				break;
			}

			iter->dma = sg_dma_address(iter->sg);
			iter->max = iter->dma + iter->sg->length;
		}

		if (++idx->pte == GEN8_PTES) {
			idx->pte = 0;

			if (++idx->pde == I915_PDES) {
				idx->pde = 0;

				/* Limited by sg length for 3lvl */
				if (++idx->pdpe == GEN8_PML4ES_PER_PML4) {
					idx->pdpe = 0;
					ret = true;
					break;
				}

				GEM_BUG_ON(idx->pdpe >= i915_pdpes_per_pdp(&ppgtt->base));
				pd = pdp->page_directory[idx->pdpe];
			}

			kunmap_atomic(vaddr);
			vaddr = kmap_atomic_px(pd->page_table[idx->pde]);
		}
	} while (1);
	kunmap_atomic(vaddr);

	return ret;
}

static void klpp_gen8_ppgtt_insert_3lvl(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   /*
				    * Fix CVE-2019-0155
				    *  -1 line, +1 line
				    */
				   u32 flags)
{
	struct i915_hw_ppgtt *ppgtt = i915_vm_to_ppgtt(vm);
	struct sgt_dma iter = sgt_dma(vma);
	struct gen8_insert_pte idx = gen8_insert_pte(vma->node.start);

	klpp_gen8_ppgtt_insert_pte_entries(ppgtt, &ppgtt->pdp, &iter, &idx,
				      /*
				       * Fix CVE-2019-0155
				       *  -1 line, +1 line
				       */
				      cache_level, flags);
}

static void klpp_gen8_ppgtt_insert_4lvl(struct i915_address_space *vm,
				   struct i915_vma *vma,
				   enum i915_cache_level cache_level,
				   /*
				    * Fix CVE-2019-0155
				    *  -1 line, +1 line
				    */
				   u32 flags)
{
	struct i915_hw_ppgtt *ppgtt = i915_vm_to_ppgtt(vm);
	struct sgt_dma iter = sgt_dma(vma);
	struct i915_page_directory_pointer **pdps = ppgtt->pml4.pdps;
	struct gen8_insert_pte idx = gen8_insert_pte(vma->node.start);

	while (klpp_gen8_ppgtt_insert_pte_entries(ppgtt, pdps[idx.pml4e++], &iter,
					     /*
					      * Fix CVE-2019-0155
					      *  -1 line, +1 line
					      */
					     &idx, cache_level, flags))
		GEM_BUG_ON(idx.pml4e >= GEN8_PML4ES_PER_PML4);
}



static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "gen8_ppgtt_insert_3lvl", (void *)&klpe_orig_gen8_ppgtt_insert_3lvl,
	  "i915" },
	{ "gen8_ppgtt_insert_4lvl", (void *)&klpe_orig_gen8_ppgtt_insert_4lvl,
	  "i915" },
};

int klp_bsc1173663_gem_gtt_resolve_kallsyms(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}

#endif
